#include <stdlib.h>
#include "codage_par_plage.h"

// Initialisation de pile
pile* creer_pile_vide() {
    pile* p = (pile*) malloc(sizeof(pile));
    p->sommet = NULL;
    p->taille = 0;
    p->nb_maillon = 0;
    return p;
}

// Déstruction de la pile
void detruire_pile(pile* p) {
    maillon* n;
    while (p->sommet) {
        n = p->sommet->suivant;
        free(p->sommet);
        p->sommet = n;
    }
    free(p);
}

void empiler(pile* p, char c) {
    maillon* m;
    if(p->sommet && p->sommet->caractere == c)
        ++p->sommet->repetition;
    else {
        m = (maillon*) malloc(sizeof(maillon));
        m->caractere = c;
        m->repetition = 0;
        m->suivant = p->sommet;
        p->sommet = m;
        ++p->nb_maillon;
    }
    ++p->taille;
}

// Prérequis : la pile est non vide
char depiler(pile* p) {
    char c = p->sommet->caractere;
    maillon* n;
    if(p->sommet->repetition == 0) {
        n = p->sommet->suivant;
        free(p->sommet);
        p->sommet = n;
        --p->nb_maillon;
    } else
        --p->sommet->repetition;
    --p->taille;
    return c;
}

pile* codage_chaine(char* s) {
    pile* encodeur = creer_pile_vide();
    for(;*s;++s)
        empiler(encodeur,*s);
    return encodeur;
}

char* decodage_chaine(pile *p) {
    char* s_decodee = (char*) malloc(sizeof(char) * p->taille + 1);
    char* tmp = s_decodee+p->taille;
    *tmp = '\0';
    while(p->sommet)
        *(--tmp) = depiler(p);
    return s_decodee;
}

char* s(uint u0, uint n) {
    uint u = u0;
    char* s = (char*) malloc(sizeof(char) * (n + 2));
    s[n+1] = '\0';
    for(uint i = 0; i <= n; i++) {
        s[i] = u % 144 == 0 ? 'N' : 'B';
        u = ( u * 15091 ) % 64007;
    }
    return s;
}

image* creer_image_noire(uint largeur, uint hauteur) {
    image* img = (image*) malloc(sizeof(image));
    img->largeur = largeur;
    img->hauteur = hauteur;
    img->pixels = (char**) malloc(sizeof(char*) * largeur);
    for(uint i = 0 ; i < img->largeur ; ++i) {
        img->pixels[i] = (char*) malloc(sizeof(char) * img->hauteur);
        for(uint j = 0 ; j < img->hauteur ; ++j)
            img->pixels[i][j] = 'N';
    }
    return img;
}

void detruire_image(image* img) {
    for(uint i = 0 ; i < img->largeur ; ++i)
        free(img->pixels[i]);
    free(img->pixels);
    free(img);
}

uint min3( uint a, uint b, uint c ) {
    if(a < b && a < c)
        return a;
    if(b < c)
        return b;
    return c;
}

uint plus_grand_carre_blanc(image* img) {
    uint max = 0;
    uint** matrice = (uint**) malloc(sizeof(uint*) * img->largeur);
    for(uint i = 0 ; i < img->largeur ; ++i) {
        matrice[i] = (uint*) malloc(sizeof(uint) * img->hauteur);
        for(uint j = 0 ; j < img->hauteur ; ++j)
            if(img->pixels[i][j] == 'B') {
                matrice[i][j] = 1;
                if(i > 0 && j > 0)
                    matrice[i][j] += min3(matrice[i-1][j],
                                          matrice[i][j-1],
                                          matrice[i-1][j-1]);
                if(matrice[i][j] > max)
                    max = matrice[i][j];
            } else {
                matrice[i][j] = 0;
            }
    }
    for(uint i = 0 ; i < img->largeur ; ++i)
        free(matrice[i]);
    free(matrice);
    return max;
}

image* i(char* s, uint t) {
    image* img = creer_image_noire(t,t);
    for(uint i = 0 ; i < img->largeur ; ++i)
        for(uint j = 0 ; j < img->largeur ; ++j)
            img->pixels[i][j] = s[i * 212 + j];
    return img;
}

uint* codage_image(image* img) {
    uint taille_code = 0;
    uint *code, *code_tmp;
    bool** coins = (bool**) malloc(sizeof(bool*) * img->largeur);
    uint** matrice = (uint**) malloc(sizeof(uint*) * img->largeur);
    for(uint i = 0 ; i < img->largeur ; ++i) {
        coins[i] = (bool*) malloc(sizeof(bool) * img->hauteur);
        matrice[i] = (uint*) malloc(sizeof(uint) * img->hauteur);
        for(uint j = 0 ; j < img->hauteur ; ++j) {
            if(img->pixels[i][j] == 'B') {
                matrice[i][j] = 1;
                coins[i][j] = true;
                ++taille_code;
                if(i > 0 && j > 0) {
                    matrice[i][j] += min3(matrice[i-1][j],
                                          matrice[i][j-1],
                                          matrice[i-1][j-1]);
                    if(coins[i-1][j] && matrice[i][j] > matrice[i-1][j]) {
                        coins[i-1][j] = false;
                        --taille_code;
                    }
                    if(coins[i][j-1] && matrice[i][j] > matrice[i][j-1]) {
                        coins[i][j-1] = false;
                        --taille_code;
                    }
                    if(coins[i-1][j-1] && matrice[i][j] > matrice[i-1][j-1]) {
                        coins[i-1][j-1] = false;
                        --taille_code;
                    }
                }
            } else {
                matrice[i][j] = 0;
                coins[i][j] = false;
            }
        }
    }
    code = (uint*) malloc(sizeof(uint) * ( taille_code * 3 + 3) );
    code_tmp = code;
    *code_tmp = taille_code;
    *(++code_tmp) = img->largeur;
    *(++code_tmp) = img->hauteur;
    for(uint i = 0 ; i < img->largeur ; ++i) {
        for(uint j = 0 ; j < img->hauteur ; ++j) {
            if(coins[i][j]) {
                *(++code_tmp) = i;
                *(++code_tmp) = j;
                *(++code_tmp) = matrice[i][j];
            }
        }
    }
    for(uint i = 0 ; i < img->largeur ; ++i) {
        free(matrice[i]);
        free(coins[i]);
    }
    free(matrice);
    free(coins);
    return code;
}

image* decodage_image(uint* code) {
    uint taille_code = code[0];
    uint largeur = code[1];
    uint hauteur = code[2];
    uint x, y, c;
    image* img = creer_image_noire(largeur,hauteur);
    for(uint k = 3 ; k <= taille_code * 3 ; k += 3) {
        x = code[k];
        y = code[k+1];
        c = code[k+2];
        for(uint i = 0; i < c ; ++i)
            for(uint j = 0; j < c ; ++j)
                img->pixels[x-i][y-j] = 'B';
    }
    return img;
}
